using System;
using System.Collections;
using System.Runtime.InteropServices;
using Microsoft.Win32;
using System.IO;

namespace NdisMonitor
{
	public class NdisHookStubs
    {
		//
		// Types.
		//

		[StructLayout( LayoutKind.Sequential, Pack = 1 )]
		public struct HOOK_STATS
		{
			public UInt32			dwLostPackets;
			public Byte				bArePacketsRemaining;
		}

		//
		// Externs.
		//

		[DllImport("kernel32.dll")]
		protected static extern IntPtr LoadLibrary(string lpFileName);

		[DllImport("NdisHook.dll", CallingConvention = CallingConvention.Cdecl)]
		protected static extern bool Start( UInt32 dwBacklog, UInt32 dwLtLoopbackIp, UInt32 dwLtNatIp, out IntPtr pplProtocolListPtr );

		[DllImport("NdisHook.dll", CallingConvention = CallingConvention.Cdecl)]
		protected static extern void Stop ();

		[DllImport("NdisHook.dll", CallingConvention = CallingConvention.Cdecl)]
		protected static extern bool Listen( UInt32[] piOaList, Int32 lOaListItems );

		[DllImport("NdisHook.dll", CallingConvention = CallingConvention.Cdecl)]
		protected static extern bool WaitForPacket( out HOOK_STATS psStats, out IntPtr pppPacketPtr );

		[DllImport("NdisHook.dll", CallingConvention = CallingConvention.Cdecl)]
		protected static extern void FreePacket( IntPtr ppPacket );

		[DllImport("NdisHook.dll", CallingConvention = CallingConvention.Cdecl)]
		protected static extern void GetHookStatsFromPacket( out HOOK_STATS psStats, IntPtr ppPacket );

		//
		// Helpers.
		//

		public class NEXT_PACKET
		{
			// # construction. #

			protected NEXT_PACKET ( IntPtr ptr, HOOK_STATS hsStats )
			{
				_ptr = ptr;
				_hsStats = hsStats;

				UInt32			dwDataLength;

				_dwOpenAdapterOrdinal = (UInt32) Marshal.ReadInt32( _ptr, 0 );
				_dwPacketSerial = (UInt32) Marshal.ReadInt32( _ptr, 4 );
				dwDataLength = (UInt32) Marshal.ReadInt32( _ptr, 12 );
				_bDirection = Marshal.ReadByte( _ptr, 16 );

				_data = new byte[ dwDataLength ];
				Marshal.Copy( new IntPtr( _ptr.ToInt32() + 17 ), _data, 0, (Int32) dwDataLength );
			}

			protected NEXT_PACKET ( HOOK_STATS hsStats, UInt32 dwOpenAdapterOrdinal, UInt32 dwPacketSerial, Byte bDirection, byte[] data )
			{
				this._hsStats = hsStats;
				this._dwOpenAdapterOrdinal = dwOpenAdapterOrdinal;
				this._dwPacketSerial = dwPacketSerial;
				this._bDirection = bDirection;
				this._data = data;
			}

			// # methods. #

			public static NEXT_PACKET WaitFor ()
			{
				HOOK_STATS	hs;
				IntPtr		ptr;

				if ( WaitForPacket( out hs, out ptr ) == false )
					return null;
				else
					return new NEXT_PACKET( ptr, hs );
			}

			public void Destroy ()
			{
				if ( _ptr != IntPtr.Zero )
					FreePacket( _ptr );
			}

			public void Serialize( BinaryWriter w )
			{
				// write state.

				w.Write( _hsStats.dwLostPackets );
				w.Write( _hsStats.bArePacketsRemaining );
				w.Write( _dwOpenAdapterOrdinal );
				w.Write( _dwPacketSerial );
				w.Write( _bDirection );
				w.Write( (UInt32) _data.Length );
				w.Write( _data );
			}

			public static NEXT_PACKET Deserialize( BinaryReader r )
			{
				// read state.

				HOOK_STATS			hsStats = new HOOK_STATS ();
				hsStats.dwLostPackets = r.ReadUInt32();
				hsStats.bArePacketsRemaining = r.ReadByte();
				UInt32				dwOpenAdapterOrdinal = r.ReadUInt32();
				UInt32				dwPacketSerial = r.ReadUInt32 ();
				Byte				bDirection = r.ReadByte();
				UInt32				dataLen = r.ReadUInt32();
				byte[]				data = r.ReadBytes( (int) dataLen );

				// return.

				return new NEXT_PACKET( hsStats, dwOpenAdapterOrdinal, dwPacketSerial, bDirection, data );
			}

			// # operator overload. #

			public static explicit operator IntPtr( NEXT_PACKET np )
			{
				if ( np._ptr == IntPtr.Zero )
					throw new Exception( "_ptr = 0;" );

				return np._ptr;
			}

			public static explicit operator NEXT_PACKET( IntPtr ptr )
			{
				if ( ptr == IntPtr.Zero )
					throw new Exception( "ptr = 0;" );

				HOOK_STATS		hs;
				GetHookStatsFromPacket( out hs, ptr );
				return new NEXT_PACKET( ptr, hs );
			}

			// # internal data. #

			protected IntPtr		_ptr = IntPtr.Zero;

			// # packet data. #

			public HOOK_STATS		_hsStats;
			public UInt32			_dwOpenAdapterOrdinal;
			public UInt32			_dwPacketSerial;
			public Byte				_bDirection;
			public byte[]			_data = null;
		}

		protected const int			_sizeOfNtProtocolList = 8;
		protected const int			_sizeOfNtProtocol = 544;
		protected const int			_sizeOfNtOpenAdapter = 584;

		public class NT_PROTOCOL_LIST
		{
			// # construction. #

			protected NT_PROTOCOL_LIST( IntPtr ptr )
			{
				_ptr = ptr;

				_dwProtocolsNum = (UInt32) Marshal.ReadInt32( _ptr, 0 );
				_dwOpenAdaptersNum = (UInt32) Marshal.ReadInt32( _ptr, 4 );

				int				i;
				int				memPos = _ptr.ToInt32() + _sizeOfNtProtocolList;

				for( i=0; i<_dwProtocolsNum; i ++ )
				{
					IntPtr		thisPtr = new IntPtr( memPos );
					_protocols.Add( new NT_PROTOCOL( thisPtr ) );
					memPos += _sizeOfNtProtocol;
				}

				for( i=0; i<_dwOpenAdaptersNum; i ++ )
				{
					IntPtr		thisPtr = new IntPtr( memPos );
					_adapters.Add( new NT_OPEN_ADAPTER( thisPtr ) );
					memPos += _sizeOfNtOpenAdapter;
				}
			}

			// # methods. #

			public static NT_PROTOCOL_LIST Start ( UInt32 backLog, UInt32 dwLtLoopbackIp, UInt32 dwLtNatIp )
			{
				// the module has to be loaded once and then it needs to remain in memory.

				if ( LoadLibrary( "NdisHook.dll" ).ToInt32() == 0 )
					return null;

				// call the Start method on the dll module.

				IntPtr		ptr;
				bool		res = NdisMonitor.NdisHookStubs.Start( backLog, dwLtLoopbackIp, dwLtNatIp, out ptr );

				if ( res == false )
					return null;
				else
					return new NT_PROTOCOL_LIST( ptr );
			}

			public void Stop ()
			{
				NdisMonitor.NdisHookStubs.Stop ();
			}

			public bool Listen( UInt32[] piOaList )
			{
				return NdisMonitor.NdisHookStubs.Listen( piOaList, ( piOaList == null ) ? 0 : piOaList.Length );
			}

			// # internal data. #

			protected IntPtr		_ptr;

			// # public data. #

			public UInt32			_dwProtocolsNum;
			public UInt32			_dwOpenAdaptersNum;
			public ArrayList		_protocols = new ArrayList ();
			public ArrayList		_adapters = new ArrayList ();
		}

		public class NT_PROTOCOL
		{
			// # construction. #

			public NT_PROTOCOL( IntPtr ptr )
			{
				_ptr = ptr;

				_dwOrdinal = (UInt32) Marshal.ReadInt32( _ptr, 0 );
				_szName = Marshal.PtrToStringUni( new IntPtr( _ptr.ToInt32() + 32 ) );
			}

			// # internal data. #

			protected IntPtr		_ptr;

			// # public data. #

			public UInt32			_dwOrdinal;
			public string			_szName;
		}

		public class NT_OPEN_ADAPTER
		{
			// # construction. #

			public NT_OPEN_ADAPTER( IntPtr ptr )
			{
				_ptr = ptr;

				// read the informations.

				_dwOrdinal = (UInt32) Marshal.ReadInt32( _ptr, 0 );
				_dwProtocolOrd = (UInt32) Marshal.ReadInt32( _ptr, 4 );
				_szName = Marshal.PtrToStringUni( new IntPtr( _ptr.ToInt32() + 16 ) );

				// try to derive the adapter name from the registry...

				adapterName = _szName;

				RegistryKey		regAdapts = Registry.LocalMachine.OpenSubKey( @"SYSTEM\CurrentControlSet\Control\Class\{4D36E972-E325-11CE-BFC1-08002BE10318}" );
				if ( regAdapts != null )
				{
					try
					{
						string[]		regSubK = regAdapts.GetSubKeyNames();
						foreach( string s in regSubK )
						{
							RegistryKey		regThis = regAdapts.OpenSubKey( s );
							if ( regThis != null )
							{
								try
								{
									object			o = regThis.GetValue( "NetCfgInstanceID" );
									if ( o != null && o is string )
									{
										// get the guid string.

										string			guid = ( (string) o ).Trim ();
										if ( guid.Length != 0 && guid[ 0 ] == '{' )
											guid = guid.Substring( 1 );
										if ( guid.Length != 0 && guid[ guid.Length - 1 ] == '}' )
											guid = guid.Substring( 0, guid.Length - 1 );

										// compare.

										if ( guid.Length == 36 &&
											_szName.IndexOf( guid ) != -1 )
										{
											object		objName = regThis.GetValue( "DriverDesc" );
											if ( objName != null && objName is string &&
												( (string) objName ).Length != 0 )
											{
												// set the name from the registry.
												adapterName = (string) objName;

												// copy also the media string.
												int		mediaPos = _szName.IndexOf( " { media" );
												if ( mediaPos != -1 )
													adapterName += _szName.Substring( mediaPos );

												// exit.
												break;
											}
										}
									}
								}
								finally
								{
									regThis.Close ();
								}
							}
						}
					}
					finally
					{
						regAdapts.Close ();
					}
				}

				// return.

				return;
			}

			// # internal data. #

			protected IntPtr		_ptr;

			// # public data. #

			public UInt32			_dwOrdinal;
			public UInt32			_dwProtocolOrd;
			public string			_szName, adapterName;
		}
	}
}
